
# This file was generated by 'versioneer.py' (0.16) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json
import sys

version_json = '''
{
 "dirty": false,
 "error": null,
 "full-revisionid": "2b78f3e1f45ff69ab529ea20f4ce1c8fcd3d2762",
 "version": "0.17.2"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
